#' @export
#'
#' @importFrom rlang .data
#'
#' @title Download Purple Air timeseries data
#'
#' @param pas Purple Air 'enhanced' synoptic data
#' @param label Purple Air 'label'
#' @param id Purple Air 'ID'
#' @param startdate Desired UTC start time (ISO 8601)
#' @param enddate Desired UTC end time (ISO 8601)
#' @param timezone Timezone used to interpret start and end dates.
#' @param baseURL Base URL for Thingspeak API
#' @return The monitor time series given broken up into metadata and readings data
#' @description Timeseries data from a specific PurpleAir can be retrieved from the Thingspeak API .
#'

downloadParseTimeseriesData <- function(
  pas = NULL,
  label = NULL,
  id = NULL,
  startdate = NULL,
  enddate = NULL,
  timezone = NULL,
  baseURL = "https://api.thingspeak.com/channels/"
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaCoreUtils::stopIfNull(pas)
  MazamaCoreUtils::stopIfNull(label)
  
  if ( !label %in% pas$label )
    stop(paste0("'", label, "' is not found in the 'pas' object"))
  
  timezone <- unique(timezone)
  
  # ----- Determine date sequence ----------------------------------------------
  
  # Subset by label
  if ( !is.null(label) ) {
    pas_single <-
      pas %>%
      dplyr::filter(.data$label == !!label)
  }
  
  # Subset by ID
  if ( !is.null(id) ) {
    pas_single <-
      pas_single %>%
      dplyr::filter(.data$ID == !!id)
  } else {
    if ( nrow(pas_single) > 1 ) {
      IDString <- paste0(sort(pas_single$ID), collapse = ", ")
      stop(paste0("Multilpe sensors share this label.",
                  "You must specify the 'id' parameter as one of: '",
                  IDString, "'"))
    }
  }
  
  # Get the timezone associated with this sensor
  if ( is.null(timezone) ) {
    timezone <-
      pas_single %>%
      dplyr::pull(.data$timezone)
  }
  
  if ( length(timezone) > 1 ) {
    err_msg <- paste0(length(timezone),
                      " senors share the same label and ID")
    stop(err_msg)
  }
  
  # Create a valid dateRange
  if ( !is.null(startdate) && !is.null(enddate) ) {
    # Don't require day boundaries
    dateRange <- MazamaCoreUtils::timeRange(startdate, 
                                            enddate, 
                                            timezone = timezone)
  } else {
    # Default to 7 days with day boundaries
    dateRange <- MazamaCoreUtils::dateRange(startdate, 
                                            enddate, 
                                            timezone, 
                                            days = 7,
                                            unit = "min")
  }
  
  startString <- strftime(dateRange[1], "%Y-%m-%dT%H:%M:%S", tz = "UTC")
  endString <- strftime(dateRange[2], "%Y-%m-%dT%H:%M:%S", tz = "UTC")
  
  # Prefer to use the monitor's ID over label
  if ( !is.null(id) ) {
    requested_meta <- dplyr::filter(pas, .data$ID == !!id)
  } else if ( !is.null(label) ) {
    requested_meta <- dplyr::filter(pas, .data$label == !!label)
  } else {
    stop("Either 'label' or 'id' must be specified.")
  }
  
  # Determine which channel was given and access the other channel from it
  if ( is.na(requested_meta$parentID) ) {
    A_meta <- requested_meta
    B_meta <- dplyr::filter(pas, .data$parentID == A_meta$ID)
  } else {
    B_meta <- requested_meta
    A_meta <- dplyr::filter(pas, .data$ID == B_meta$parentID)
  }
  
  # Combine channel A and B monitor metadata
  meta <- dplyr::bind_rows(A_meta, B_meta)
  
  # Generate Thingspeak request URLs
  A_url <- 
    paste0(
      baseURL,
      A_meta$THINGSPEAK_PRIMARY_ID,
      "/feeds.json?api_key=",
      A_meta$THINGSPEAK_PRIMARY_ID_READ_KEY,
      "&start=", startString,
      "&end=",
      endString
    )
  B_url <-
    paste0(
      baseURL,
      B_meta$THINGSPEAK_PRIMARY_ID,
      "/feeds.json?api_key=",
      B_meta$THINGSPEAK_PRIMARY_ID_READ_KEY,
      "&start=", startString,
      "&end=",
      endString
    )
  
  # ----- Request A channel data from Thingspeak -------------------------------
  
  webserviceUrl <- A_url
  
  # message(webserviceUrl)
  
  # NOTE:  using Hadley Wickham style:
  # NOTE:  https://github.com/hadley/httr/blob/master/vignettes/quickstart.Rmd
  r <- httr::GET(webserviceUrl)
  
  # Handle the response
  status_code <- httr::status_code(r)
  content <- httr::content(r, as = "text") # don't interpret the JSON
  
  # Handle "no data" response by generating an empty but complete "pat" object
  err_JSON <- ('{
       "channel": {
        "id": 0,
        "name": "NA",
        "latitude": "0.0",
        "longitude": "0.0",
        "field1": "PM1.0 (ATM)",
        "field2": "PM2.5 (ATM)",
        "field3": "PM10.0 (ATM)",
        "field4": "Uptime",
        "field5": "RSSI",
        "field6": "Temperature",
        "field7": "Humidity",
        "field8": "PM2.5 (CF=1)",
        "created_at": "2000-01-01T12:00:00Z",
        "updated_at": "2000-01-01T12:00:00Z",
        "last_entry_id": 0
      },
      "feeds": [
        {
          "created_at": "2000-01-01T12:00:00Z",
          "entry_id": 0,
          "field1": "NA",
          "field2": "NA",
          "field3": "NA",
          "field4": "NA",
          "field5": "NA",
          "field6": "NA",
          "field7": "NA",
          "field8": "NA"
        }
        ]
      }')
  
  err_list <- jsonlite::fromJSON(
    txt = err_JSON,
    simplifyVector = TRUE,
    simplifyDataFrame = TRUE,
    simplifyMatrix = TRUE,
    flatten = FALSE
  ) 

  err_data <- err_list$feeds
  
  if ( httr::http_error(r) ) { # web service failed to respond
    
    # https://digitalocean.com/community/tutorials/how-to-troubleshoot-common-http-error-codes
    if ( httr::status_code(r) == 429 ) {
      err_msg <- paste0(
        "web service error 429: Too Many Requests from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 500 ) {
      err_msg <- paste0(
        "web service error 500: Internal Server Error from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 502 ) {
      err_msg <- paste0(
        "web service error 502: Bad Gateway from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 503 ) {
      err_msg <- paste0(
        "web service error 503: Service Unavailable from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 504 ) {
      err_msg <- paste0(
        "web service error 504: Gateway Timeout from ",
        webserviceUrl
      )
    } else {
      err_msg <- paste0(
        "web service error ", httr::status_code(r), " from ",
        webserviceUrl
      )
    }
    
    message(paste0("Channel A: ", err_msg, " === Returning empty A channel ==="))
    
    A_list <- err_list
    A_data <- err_data
    
  } else { # Response successful
    
    A_list <- 
      jsonlite::fromJSON(
        content,
        simplifyVector = TRUE,
        simplifyDataFrame = TRUE,
        simplifyMatrix = TRUE,
        flatten = FALSE
      )
    A_data <- A_list$feeds
    
  }
  
  # ----- Request B channel data from Thingspeak -------------------------------
  
  webserviceUrl <- B_url
  
  # NOTE:  using Hadley Wickham style:
  # NOTE:  https://github.com/hadley/httr/blob/master/vignettes/quickstart.Rmd
  r <- httr::GET(webserviceUrl)
  
  # Handle the response
  status_code <- httr::status_code(r)
  content <- httr::content(r, as = "text") # don't interpret the JSONw
  
  if ( httr::http_error(r) ) { # web service failed to respond
    
    # https://digitalocean.com/community/tutorials/how-to-troubleshoot-common-http-error-codes
    if ( httr::status_code(r) == 429 ) {
      err_msg <- paste0(
        "web service error 429: Too Many Requests from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 500 ) {
      err_msg <- paste0(
        "web service error 500: Internal Server Error from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 502 ) {
      err_msg <- paste0(
        "web service error 502: Bad Gateway from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 503 ) {
      err_msg <- paste0(
        "web service error 503: Service Unavailable from ",
        webserviceUrl
      )
    } else if ( httr::status_code(r) == 504 ) {
      err_msg <- paste0(
        "web service error 504: Gateway Timeout from ",
        webserviceUrl
      )
    } else {
      err_msg <- paste0(
        "web service error ", httr::status_code(r), " from ",
        webserviceUrl
      )
    }
    
    message(paste0("Channel B: ", err_msg, " === Returning empty B channel ==="))
    
    B_list <- err_list
    B_data <- err_data
    
  } else {
    
    # Response successful
    
    B_list <- 
      jsonlite::fromJSON(
        content,
        simplifyVector = TRUE,
        simplifyDataFrame = TRUE,
        simplifyMatrix = TRUE,
        flatten = FALSE
      )
    B_data <- B_list$feeds
    
  }
  
  # Sanity check for data -> fill if empty to avoid error DL 
  if ( length(A_data) == 0 && length(B_data) == 0 ) {
    A_data <- err_data
    B_data <- err_data
    warning(
     paste0(label,": A & B channels for the requested time period do not exist.")
    )
  } else if ( length(A_data) == 0) {
    A_data <- err_data
    warning(
      paste0(label,": A channel for the requested time period do not exist.")
    )
  } else if ( length(B_data) == 0) {
    B_data <- err_data
    warning(
      paste0(label,": B channel for the requested time period do not exist")
    )
  }
  
  # Rename columns
  names(A_data) <- c(
    "datetime", "entry_id", "pm1_atm", "pm2.5_atm", "pm10_atm",
    "uptime", "rssi", "temperature", "humidity", "pm2.5_cf1"
  )
  names(B_data) <- c(
    "datetime", "entry_id", "pm1_atm", "pm2.5_atm", "pm10_atm",
    "memory", "adc0", "unused1", "unused2", "pm2.5_cf1"
  )
  
  # Add channel identifier
  A_data$channel <- "A"
  B_data$channel <- "B"
  
  # Drop unused columns
  B_data <- dplyr::select(B_data, -.data$unused1, -.data$unused2)
  
  # Combine data from both channels
  data <-
    dplyr::bind_rows(A_data, B_data) %>%
    dplyr::arrange(.data$datetime)
  
  # NOTE:  > names(data)
  # NOTE:  [1] "datetime"    "entry_id"    "pm1_atm"     "pm2.5_atm"   "pm10_atm"    "uptime"      "rssi"
  # NOTE:  [8] "temperature" "humidity"    "pm2.5_cf1"   "channel"     "memory"      "adc0"
  
  numeric_columns <- c(
    "pm1_atm", "pm2.5_atm", "pm10_atm", "uptime", "rssi",
    "temperature", "humidity", "pm2.5_cf1", "memory", "adc0"
  )
  
  # Convert to proper types
  data$datetime <- lubridate::ymd_hms(data$datetime, tz="UTC")
  data$entry_id <- as.character(data$entry_id)
  for (columnName in numeric_columns) {
    data[[columnName]] <- as.numeric(data[[columnName]])
  }
  
  # Round values to reflect resolution as specified in
  # https://www.purpleair.com/sensors
  
  # NOTE:  Rounding breaks outlier detection
  # data$pm1_atm <- round(data$pm1_atm)
  # data$pm1_atm <- round(data$pm1_atm)
  # data$pm2.5_atm <- round(data$pm2.5_atm)
  # data$pm10_atm <- round(data$pm10_atm)
  # data$temperature <- round(data$temperature)
  # data$humidity <- round(data$humidity)
  # data$pm2.5_cf1 <- round(data$pm2.5_cf1)
  
  # Combine meta and data dataframes into a list
  pat_raw <- list(meta = meta, data = data)
  
  # Remove any duplicate data records
  pat_raw$data <- dplyr::distinct(pat_raw$data)
  
  # ----- Return ---------------------------------------------------------------
  
  return(pat_raw)
  
}

# TODO:  Probably have an internal function to create an empty "pat" object.
# TODO:  This can then be returned immediately whenever a "no data" response
# TODO:  is detected when requesting data.

# ===== DEBUGGING ==============================================================

if ( FALSE ) {
  
  pat_raw <- downloadParseTimeseriesData(
    pas = pas_load(),
    label = "SCAP_28",
    id = NULL,
    startdate = 20190701,
    enddate = 20190805,
    baseURL = "https://api.thingspeak.com/channels/"
  )
  
  pat <- createPATimeseriesObject(pat_raw)
  

}
